/*********************************************************************
 *
 *                Microchip USB C18 Firmware - MCP3551 PICtail Demo
 *
 *********************************************************************
 * FileName:        PowerMeter.c
 * Dependencies:    See INCLUDES section below
 * Processor:       PIC18
 * Compiler:        C18 2.30.01+
 * Company:         Microchip Technology, Inc.
 *
 * Software License Agreement
 *
 * The software supplied herewith by Microchip Technology Incorporated
 * (the Company) for its PICmicro Microcontroller is intended and
 * supplied to you, the Companys customer, for use solely and
 * exclusively on Microchip PICmicro Microcontroller products. The
 * software is owned by the Company and/or its supplier, and is
 * protected under applicable copyright laws. All rights are reserved.
 * Any use in violation of the foregoing restrictions may subject the
 * user to criminal sanctions under applicable laws, as well as to
 * civil liability for the breach of the terms and conditions of this
 * license.
 *
 * THIS SOFTWARE IS PROVIDED IN AN AS IS CONDITION. NO WARRANTIES,
 * WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED
 * TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. THE COMPANY SHALL NOT,
 * IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL OR
 * CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
 *
 * Author               Date        Comment
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Pat Richards       xx/xx/xx    Original.
 * Jerry Horn         01/20/06    Changed from SPI master to I2C slave device
 * Jerry Horn         06/15/06    Changes to run on Power Meter USB Interface Module
 ********************************************************************/

/** I N C L U D E S **********************************************************/
#include <p18cxxx.h>
#include <usart.h>
#include <i2c.h>
#include "system\typedefs.h"
#include "system\usb\usb.h"

#include "io_cfg.h"             // I/O pin mapping
#include "user\PowerMeter.h"

/** V A R I A B L E S ********************************************************/
#pragma udata

BOOL txReady;
BOOL gSampleFlag;

BOOL I2C_xfer_done;
unsigned char I2C_temp[3];
unsigned char I2C_result[3];
unsigned char I2C_counter;

char inbuffer[BUF_SIZE];		// 64 byte input to USB device buffer
char outbuffer[BUF_SIZE];		// 64 byte output to USB device buffer
byte TimerCounter;


/** P R I V A T E  P R O T O T Y P E S ***************************************/
void Read3551(char *data);
unsigned char ReadSPI( void );
void CheckBoardConnect(void);

/** D E C L A R A T I O N S **************************************************/
#pragma code
void UserInit(void)
{
  byte i;
  
  //Input/output pins
  tris_CS = 1; 					//Input, not used for I2C
  TRISBbits.TRISB0 = 1;   		//SDA
  TRISBbits.TRISB1 = 1;   		//SCL

  //Setup I2C registers
  I2C_xfer_done = FALSE;
  SSPADD = 0xC0;				//I2C address of this micro
  PIE1bits.SSPIE = 0;			//Turn off I2C/SPI interrupt
  PIR1bits.SSPIF = 0;			//Clear any pending interrupt
  OpenI2C(SLAVE_7,SLEW_ON);
  PIE1bits.SSPIE = 1;			//Turn on I2C interrupt

  //-------------------------
  // initialize variables
  //-------------------------
  for (i=0; i<BUF_SIZE; i++)     // initialize input and output buffer to 0
  {
    inbuffer[i] = 0;
    outbuffer[i] = 0;
  }
  gSampleFlag = FALSE;
  txReady = FALSE;
  TimerCounter = 0xF0;			//Initial value to start
    
  //Timer 0
  TMR0H = 0;					//clear timer
  TMR0L = 0;					//clear timer
  T0CONbits.PSA = 0;			//Assign prescaler to Timer 0
  T0CONbits.T0PS2 = 1;			//Setup prescaler
  T0CONbits.T0PS1 = 1;			//Will time out every 51 us based on
  T0CONbits.T0PS0 = 1;			//20 MHz Fosc
  T0CONbits.T0CS = 0;			//Increment on instuction cycle

  //Enable I2C interrupts
  INTCONbits.PEIE = 1;			//Turn on peripheral interrupts
  INTCONbits.GIE = 1;			//Turn on global interrupts

}//end UserInit

/******************************************************************************
 * Function:        void ProcessIO(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        This function is a place holder for other user routines.
 *                  It is a mixture of both USB and non-USB tasks.
 *
 * Note:            None
 *****************************************************************************/
void ProcessIO(void)
{   
  char n;
  byte len;
  byte i;
  byte j;
  byte readSize;

  // User Application USB tasks
  if((usb_device_state < CONFIGURED_STATE)||(UCONbits.SUSPND==1)) return;
    
  if (len = HIDRxReport(inbuffer, 64))	// USB receive buffer has data
  {
	if (len == 64)
    {
      switch(inbuffer[0])				// interpret command
      {
        case CAL_INIT:				// 'C' initialize calibration registers
        case ECHO_DATA:				// 'E' echo received data
        case LOAD_CAL_FROM_FLASH:		// 'L' load calibration registers from flash
        case STORE_CAL_TO_FLASH:		// 'S' store calibration registers to flash
        case WRITE_DATA_TO_ADDR:		// 'W' load calibration registers from flash
          for (i = 0; i < len; i++)
	        outbuffer[i] = inbuffer[i];
          txReady = TRUE;
        break;

        case READ_DATA_FROM_ADDR:		// 'R' load calibration registers from flash
		  for (i = 0; i < 4; i++)
	        outbuffer[i] = inbuffer[i];
		  inbuffer[4] = inbuffer[4] | 32;  //Convert hex alpha to lowercase
          inbuffer[5] = inbuffer[5] | 32;  //Convert hex alpha to lowercase
	      readSize = 0;
          if ((inbuffer[4] >= '0') && (inbuffer[4] <= '9'))
            readSize = (inbuffer[4] - '0') * 16;
          else if ((inbuffer[4] >= 'a') && (inbuffer[4] <= 'f'))
            readSize = (inbuffer[4] - 'a') * 16;
          if ((inbuffer[5] >= '0') && (inbuffer[5] <= '9'))
            readSize = readSize + (inbuffer[5] - '0');
          else if ((inbuffer[5] >= 'a') && (inbuffer[5] <= 'f'))
            readSize = readSize + (inbuffer[5] - 'a');
		  if ((readSize >= 1) && (readSize <= 29))
		    for (i = 0; i < readSize; i++)
            {
			  j = i << 1;
	          outbuffer[4 + j] = j;
	          outbuffer[4 + j + 1] = j + 1;
            }
		  readSize = readSize << 1;
          outbuffer[4 + readSize] = TERM;
		  txReady = TRUE;
        break;

        default:						// unrecognized or null command
	      outbuffer[0] = TERM;
		  for (i = 1; i < len; i++)
	        outbuffer[i] = 0;
		  txReady = TRUE;
      }// END switch(inbuffer[0])
    }// if (len == 64)
    else
    {
      outbuffer[0] = TERM;
      txReady = TRUE;
    }
  }//END if (len = HIDRxReport(inbuffer, 64))

  //Inst. cycle = 200 ns; TMR0IF sets every 51 us
  if(INTCONbits.TMR0IF)	
  {
  	TimerCounter++;
  	if (!TimerCounter)			//if rolled over, set flag. User code will handle the rest.
  	{
  	  TimerCounter = 0xFE;
  	  gSampleFlag = TRUE;
  	}
	else
	  gSampleFlag = FALSE;
  	INTCONbits.TMR0IF = 0;
  }
  
  //MCU checks every 100 us if conversion is finished
  //The commented if only sends data if all conditions are met.
  //However, this results in Dataview reporting 0. Not sure why,
  //but Dataview may not be checking for valid data. Basically, it
  //appears that old data must be re-sent in order to keep either
  //Dataview happy or for some USB HID reason.

  //if(gSampleFlag && I2C_xfer_done && !mHIDTxIsBusy())
  if(txReady && gSampleFlag && !mHIDTxIsBusy())
  {
	//if (I2C_xfer_done)			//this if should really not be here
	//  Read3551(outbuffer);
   HIDTxReport(outbuffer,64);	//this keeps dataview happy by sending old data (in some cases)
   txReady = FALSE;
  }

}//end ProcessIO

/******************************************************************************
 * Function:        void Read3551(char *data)
 *
 * PreCondition:    None
 *
 * Input:           Pointer to a string; must be three bytes min
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        
 *
 * Note:            
 *****************************************************************************/
void Read3551(char *data)
{
  //Need an atomic operation here
  PIE1bits.SSPIE = 0;		//Turn I2C interrupts off
  data[2] = I2C_result[2];  
  data[1] = I2C_result[1];  
  data[0] = I2C_result[0];
  I2C_xfer_done = FALSE;
  PIE1bits.SSPIE = 1;		//Turn I2C interrupts on
}

/******************************************************************************
 * Function:        CheckBoardConnect(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Overview:        Checks if the Stimulus Board is attached. Future expansion.        
 *
 * Side Effects:    None
/******************************************************************************/
void CheckBoardConnect(void)
{
}

/** EOF MCP3551.c********************************************************/
